@@ -13,9 +13,11 @@ sub DUMP {
     print STDERR YAML::Dump(@_);
 }
 sub PRINT {
-    chomp $_[-1];
     print STDERR @_, "\n";
 }
+sub DIE {
+    die @_, "\n";
+}
 
 ###############################################################################
 # command support routines
@@ -41,12 +43,38 @@ sub break {
     print STDOUT ".saves breakpoint created\n";
 }
 
+sub delete {
+    my ($self, $switches, $files) = @_;
+    assert_repository();
+    validate_files($files);
+    my %manifest = map {($_, 1)} @{read_manifest()};
+    my $file_list = files_in_manifest($files);
+    DIE "None of these files are in the manifest"
+      unless @$file_list;
+    if (not defined $switches->{f}) {
+        printf "Are you sure you want to delete %d files from the repository? ",
+               0+@$file_list;
+        my $answer = <STDIN>;
+        unless ($answer =~ /^(y|yes)$/i) {
+            print STDOUT "svs delete aborted\n";
+            return;
+        }
+    }
+    for (@$file_list) {
+        delete $manifest{$_};
+        system("rm -f .saves/SAVES/$_,v") == 0
+          or DIE "Can't delete .saves/SAVES/$_,v";
+    }
+    write_manifest([keys %manifest]);
+    printf STDOUT "%d files deleted from .saves repository\n", 0+@$file_list;
+}
+
 sub diff {
     my ($self, $switches, $files) = @_;
     assert_repository();
     push @$files, '.' unless @$files;
     validate_files($files);
-    show_diff($files);
+    show_diff($files, $switches);
 }
 
 sub find {
@@ -61,7 +89,7 @@ sub find {
 sub help {
     my ($self, $switches, $sections) = @_;
     my $section;
-    die "'svs help' only takes one argument at a time\n"
+    DIE "'svs help' only takes one argument at a time"
       if @$sections > 1;
     require VCS::SaVeS::Help;
     $section = @$sections
@@ -119,15 +147,12 @@ sub restore {
     assert_repository();
     assert_file_in_repository($files);
     my $file = $files->[0];
-    my $revision = $switches->{r} || 1;
-    if ($revision =~ /^\d+\./) {
-        restore_by_revision($file, $revision);
-    }
-    elsif ($revision =~ /^\d+$/ and $revision > 0) {
+    my $revision = $switches->{r} || @{parse_rlog($file)};
+    if ($revision =~ /^\d+$/ and $revision > 0) {
         restore_by_number($file, $revision);
     }
     else{
-        die "-n must be positive integer or revision number\n";
+        DIE "-n must be positive integer";
     }
 }
 
@@ -162,7 +187,7 @@ VERSION
 ###############################################################################
 sub get_message {
     my ($switches) = @_;
-    my $default = 'No worries mate';
+    my $default = '...';
     return $default if defined $switches->{M};
 
     if (defined $switches->{m}) {
@@ -195,7 +220,7 @@ sub write_manifest {
 
     chmod 0644, '.saves/MANIFEST';
     open MANIFEST, "> .saves/MANIFEST" 
-        or die $!;
+        or DIE $!;
     print MANIFEST <<END;
 #==============================================================================
 #
@@ -238,7 +263,7 @@ sub initialize_repository {
               qq{rcs -q -i .saves/SAVES/$file,v < /dev/null\n};
         }
     }
-    open SH, "| sh" or die $!;
+    open SH, "| sh" or DIE $!;
     print SH $shell_commands;
     close SH;
 }
@@ -246,19 +271,30 @@ sub initialize_repository {
 sub update_repository {
     my ($files, $msg) = @_;
     my $shell_commands = '';
+    my %stamp;
     for my $file (@$files) {
         $shell_commands .= qq{ci -q -l -m"$msg" $file .saves/SAVES/$file,v\n};
+        $stamp{$file} = -M ".saves/SAVES/$file,v";
     }
-    open SH, "| sh" or die $!;
+    open SH, "| sh" or DIE $!;
     print SH $shell_commands;
     close SH;
-    return scalar @$files;
+    my $count = 0;
+    for my $file (@$files) {
+        $count++, next if $stamp{$file} < .000001;
+        next unless -e ".saves/SAVES/$file,v";
+        if ((-M $file) < (-M ".saves/SAVES/$file,v")) {
+            system "touch .saves/SAVES/$file,v";
+        }
+        $count++ if (-M ".saves/SAVES/$file,v") < $stamp{$file};
+    }
+    return $count;
 }
 
 sub read_manifest {
     my $files = [];
     open MANIFEST, '< .saves/MANIFEST'
-      or die "Can't open .saves/MANIFEST for input\n";
+      or DIE "Can't open .saves/MANIFEST for input";
     @$files = map {chomp; $_} grep {not /^\s*\#/} <MANIFEST>;
     close MANIFEST;
     return $files;
@@ -309,14 +345,14 @@ sub find_files {
         return () if -e "$file/.saves" and $file !~ /^\.\/?$/;
         my @files = ();
         local *DIR;
-        opendir(DIR, $file) or die "Can't opendir $file";
+        opendir(DIR, $file) or DIE "Can't opendir $file";
         while (my $new_file = readdir(DIR)) {
             next if $new_file =~ /^(\.|\.\.|(\.\/)\.saves)$/;
             push @files, find_files($new_file, $file);
         }
         return @files;
     }
-    die "Don't know how to handle $file\n";
+    DIE "Don't know how to handle $file";
 }
 
 sub files_in_manifest {
@@ -339,11 +375,11 @@ sub show_status {
     my ($file_list) = @_;
     mkdir(".saves/tmp", 0777) unless -d ".saves/tmp";
     open STATUSLIST, "> .saves/tmp/statuslist"
-      or die $!;
+      or DIE $!;
     print STATUSLIST ".saves/SAVES/$_,v\n" for @$file_list;
     close STATUSLIST;
     open STATUSTEXT, "cat .saves/tmp/statuslist | xargs rlog -zLT |"
-      or die $!;
+      or DIE $!;
     local $/;
     my $statustext = <STATUSTEXT>;
     close STATUSTEXT;
@@ -352,16 +388,17 @@ sub show_status {
     for my $section (@sections) {
         $section =~ 
           /^RCS file: (.*?)\n.*?^locks:.*?:\s+(.*?)\n/sm
-            or die "Can't grok rlog output:\n$section\n";
+            or DIE "Can't grok rlog output:\n$section";
         my ($version, $file) = ($2, $1);
         $section =~ 
           /^revision\s+\Q$version\E.*?\n.*?date:\s+(.*?);/sm
-            or die "Can't grok rlog output:\n$section\n";
+            or DIE "Can't grok rlog output:\n$section";
         my $date = $1;
         $file =~ s{^\.saves/SAVES/(.*),v$}{$1};
         my $modified = ((-M $file) < (-M ".saves/SAVES/$file,v"))
                        ? '*'
                        : ' ';
+        $version =~ s/^\d+\.//;
         print STDOUT "$date ($version)$modified$file\n";
     }
 }
@@ -378,19 +415,21 @@ sub dump_history {
         
         chomp $_->{message};
         printf STDOUT 
-          "%d) %s (%s %6s) %s\n",
+          "%d) %s (%6s) %s\n",
           $i++, 
           $_->{date},
-          $_->{revision},
           $_->{delta},
           $message;
     }
+    if ((-M $file) < (-M ".saves/SAVES/$file,v")) {
+        print STDOUT "*) Working file has been modified since last save\n";
+    }
 }
 
 sub parse_rlog {
     my ($file) = @_;
     open RLOG, "rlog -zLT .saves/SAVES/$file |"
-      or die $!;
+      or DIE $!;
     local $/;
     my $input = <RLOG>;
     close RLOG;
@@ -398,13 +437,13 @@ sub parse_rlog {
     my @rlog = split /^-+\n/m, $rlog;
     shift(@rlog);
     my $parse;
-    for (@rlog) {
+    for (reverse @rlog) {
         /^revision\s+(\S+).*?
          ^date:\s+(.+?);.*?
          (?:lines:\s+(.+?))?\n
          (?:branches:.*?\n)?
          (.*)
-        /xms or die "Couldn't parse rlog for '$file':\n$rlog";
+        /xms or DIE "Couldn't parse rlog for '$file':\n$rlog";
         push @$parse,
           {
             revision => $1,
@@ -418,42 +457,70 @@ sub parse_rlog {
 
 sub restore_by_revision {
     my ($file, $revision) = @_;
-    my %revisions = map {($_->{revision}, 1)} @{parse_rlog($file)};
-    die "Revision number is invalid\n"
+    my @revisions = @{parse_rlog($file)};
+    my %revisions = map {($_->{revision}, 1)} @revisions;
+    my $latest = $revisions[-1]{revision};
+    DIE "Revision number is invalid"
       unless defined $revisions{$revision};
-    system(qq{rcs -q -u $file .saves/SAVES/$file,v && } .
-           qq{co -q -f -l$revision $file .saves/SAVES/$file,v}) == 0
-      or die "Couldn't restore file '$file', revision '$revision'\n";
+    my $command = join ' && ',
+      qq{rcs -q -u $file .saves/SAVES/$file,v},
+      qq{co -q -f$revision $file .saves/SAVES/$file,v},
+      qq{rcs -q -l$latest $file .saves/SAVES/$file,v},
+      qq{chmod +w $file},
+      qq{sleep 1},
+      qq{touch $file};
+    system($command) == 0
+      or DIE "Couldn't restore file '$file', revision '$revision'";
 }
 
 sub restore_by_number {
     my ($file, $number) = @_;
     my $rlog = parse_rlog($file);
-    die "Revision number is invalid\n"
+    DIE "Revision number is invalid"
       if $number-- > @$rlog;
     my $revision = $rlog->[$number]{revision};
-    system(qq{rcs -q -u $file .saves/SAVES/$file,v && } .
-           qq{co -q -f -l$revision $file .saves/SAVES/$file,v}) == 0
-      or die "Couldn't restore file '$file', revision '$revision'\n";
+    restore_by_revision($file, $revision);
 }
 
 sub show_diff {
-    my ($files) = @_;
+    my ($files, $switches) = @_;
+    my @files = @{files_in_manifest($files)};
+    my $options = '';
+    if (defined $switches->{r}) {
+        DIE "'diff -r' can only be used on a single file"
+          unless (@files == 1);
+        DIE "Invalid value for -r"
+          unless $switches->{r} =~ /^(\d+)(?:-(\d+))?$/;
+        my ($r1, $r2) = ($1, $2);
+        my $rlog;
+        for ($r1, $r2) {
+            next unless defined;
+            unless (/\./) {
+                $_ = $_ - 1;
+                $rlog ||= parse_rlog($files[0]);
+                DIE "Revision number is invalid"
+                  if $_ >= @$rlog;
+                $_ = $rlog->[$_]{revision};
+            }
+        }
+        $options = "-r$r1";
+        $options .= " -r$r2" if defined($r2);
+    }
     mkdir(".saves/tmp", 0777) unless -d ".saves/tmp";
     if (-e '.saves/tmp/diff') {
         unlink('.saves/tmp/diff')
-          or die "Can't unlink .saves/tmp/diff";
+          or DIE "Can't unlink .saves/tmp/diff";
     }
     my $shell_commands;
-    for my $file (@{files_in_manifest($files)}) {
+    for my $file (@files) {
         $shell_commands .= 
-          qq{rcsdiff -q -zLT -u $file .saves/SAVES/$file,v} .
+          qq{rcsdiff -q -zLT -u $options $file .saves/SAVES/$file,v} .
           qq{ >> .saves/tmp/diff\n};
     }
-    open SH, "| sh" or die $!;
+    open SH, "| sh" or DIE $!;
     print SH $shell_commands;
     close SH;
-    open DIFF, '.saves/tmp/diff' or die $!;
+    open DIFF, '.saves/tmp/diff' or DIE $!;
     local $/;
     print STDOUT <DIFF>;
     close DIFF;
@@ -464,7 +531,7 @@ sub show_diff {
 ###############################################################################
 sub assert_repository {
     (my $command = (caller(1))[3]) =~ s/.*::(\w+?)_?$/$1/;
-    die "Can't do 'svs $command'; no repository in this directory.\n",
+    DIE "Can't do 'svs $command'; no repository in this directory.\n",
         "You can use 'svs import' to create a repository.\n"
           unless -d ".saves";
 }
@@ -472,7 +539,7 @@ sub assert_repository {
 sub assert_no_repository {
     if (-d '.saves') {
         (my $command = (caller(1))[3]) =~ s/.*::(\w+?)_?$/$1/;
-        die <<END;
+        DIE <<END;
 Can't do 'svs $command'; a '.saves' repository already exists.
 If you really want to $command, you must first remove the repository.
 END
@@ -481,7 +548,7 @@ END
 
 sub validate_no_files {
     my ($files) = @_;
-    die "You can't specify files for this command\n"
+    DIE "You can't specify files for this command\n"
       if @$files;
 }
 
@@ -489,15 +556,15 @@ sub validate_files {
     my ($files) = @_;
     my %paths;
     (my $command = (caller(1))[3]) =~ s/.*::(\w+?)_?$/$1/;
-    die "No files specified for 'svs $command'\n"
+    DIE "No files specified for 'svs $command'\n"
       unless @$files;
     for (@$files) {
         my $file = $_;
-        die "Absolute pathnames may not be used\n"
+        DIE "Absolute pathnames may not be used\n"
           if $file =~ /^[\/\\]/;
-        die "Paths containing '../' not allowed\n"
+        DIE "Paths containing '../' not allowed\n"
           if $file =~ /\.\.\//;
-        die "$file does not exist\n"
+        DIE "$file does not exist\n"
           unless -e $file;
         if ($file =~ m|/| or -d $file) {
             my @dirs = split '/', $file;
@@ -512,7 +579,7 @@ sub validate_files {
     }
     delete $paths{'.'};
     for my $path (sort keys %paths) {
-        die <<END if -d "$path/.saves";
+        DIE <<END if -d "$path/.saves";
 Can't use files in '$path'. 
 It contains its own '.saves' directory.
 Use 'svs merge $path',
@@ -524,14 +591,14 @@ END
 sub assert_file_in_repository {
     my ($files) = @_;
     (my $command = (caller(1))[3]) =~ s/.*::(\w+?)_?$/$1/;
-    die "'svs $command' requires one filename\n"
+    DIE "'svs $command' requires one filename\n"
       unless @$files == 1;
     $files->[0] =~ s|^\./||;
     my $file = $files->[0];
-    die "'$file' is not a regular file\n"
+    DIE "'$file' is not a regular file\n"
       unless -f $file;
     my %manifest = map {($_, 1)} @{read_manifest()};
-    die "'$file' is not in the manifest\n"
+    DIE "'$file' is not in the manifest\n"
       unless $manifest{$file};
 }
 
